@@ -100,7 +100,7 @@ Details:
 
     print $tree->{family}->{father};        # the father's given name.
 
-The prefix '-' is added on every attributes' name.
+The prefix '-' is added on every attribute's name.
 
     print $tree->{family}->{"-name"};       # the family name of the family
 
@@ -113,7 +113,7 @@ The array is used because the family has two boys.
 
 If a element has both of a text node and attributes
 or both of a text node and other child nodes,
-value of a text node is moved to C<'#text'> like child nodes.
+value of a text node is moved to C<#text> like child nodes.
 
     use XML::TreePP;
     use Data::Dumper;
@@ -128,144 +128,266 @@ The result dumped is following:
     $VAR1 = {
         'span' => {
             '-class' => 'author',
-            '#text' => 'Kawasaki Yusuke'
+            '#text'  => 'Kawasaki Yusuke'
         }
     };
 
-The special node name of C<'#text'> is used because this elements
+The special node name of C<#text> is used because this elements
 has attribute(s) in addition to the text node.
-
-=head1 CONSTRUCTOR AND OPTIONS
-
-=head2 $tpp = XML::TreePP->new();
-
-This constructor method returns a new XML::TreePP object.
-
-=head2 $tpp = XML::TreePP->new( %options );
-
-Its first argument is a hash variable to set one or more options
-like following:
-
-=head2 $tpp->set( option_name => $option_value );
-
-This method sets a option value for "option_name".
-If $option_value is not defined, its option is deleted.
-Options below are available:
-
-=head2 $tpp->set( output_encoding => 'UTF-8' );
-
-You can define a encoding of xml file generated by write/writefile
-methods. On Perl 5.8.x and later, you can select it from every
-encodings supported by Encode.pm. On Perl 5.6.x or before with
-Jcode.pm, you can use 'Shift_JIS', 'EUC-JP', 'ISO-2022-JP' and
-'UTF-8'. The default value is 'UTF-8'.
-
-=head2 $tpp->set( force_array => [ 'rdf:li', 'item', '-xmlns' ] );
-
-This option allows you to specify a list of element names which
-should always be forced into an array representation
-The default value is null, it means that context of the elements
-will determine to make array or to keep it scalar.
-
-=head2 $tpp->set( first_out => [ 'link', 'title', '-type' ] );
-
-This option allows you to specify a list of element/attribute
-names which should always appears at first on output XML code.
-The default value is null, it means alphabetical order is used.
-
-=head2 $tpp->set( last_out => [ 'items', 'item', 'entry' ] );
-
-This option allows you to specify a list of element/attribute
-names which should always appears at last on output XML code.
-
-=head2 $tpp->set( cdata_scalar_ref => 1 );
-
-This option allows you to convert a cdata section into a reference
-for scalar on parsing XML source. If this option is false, per
-default, cdata section is converted into a scalar.
-
-=head2 $tpp->set( user_agent => 'Mozilla/4.0 (compatible; ...)' );
-
-This option allows you to specify a HTTP_USER_AGENT string which
-is used by parsehttp() method.
-The default string is C<"XML-TreePP/#.##">, where C<"#.##"> is
-substituted with the version number of this library.
-
-=head2 $tpp->set( attr_prefix => '@' );
-
-This option allows you to specify a prefix character(s) which
-is inserted before each attribute names.
-The default character is C<'-'>.
-Or set C<'@'> to access attribute values like E4X, ECMAScript for XML.
-Zero-length prefix C<''> is also available now.
-
-=head2 $tpp->set( text_node_key => '#text' );
-
-This option allows you to specify a hash key for text nodes.
-The default key is C<'#text'>.
-
-=head2 $tpp->set( ignore_error => 1 );
-
-This module calls Carp::croak function on an error per default.
-This option makes all errors ignored and just return.
-
-=head2 $tpp->set( xml_decl => '' );
-
-This module generates an XML declaration on writing an XML code per default.
-This option forces to change or leave it.
-
-=head2 $tpp->set( http_lite => $http );
-
-This option forces pasrsehttp() method to use L<HTTP::Lite> module 
-with its instance created like: C<$http = HTTP::Lite-E<gt>new();>
-
-=head2 $tpp->set( lwp_useragent => $ua );
-
-This option forces pasrsehttp() method to use L<LWP::UserAgent> module 
-with its instance created like: C<$ua = LWP::UserAgent-E<gt>new();>
-
-=head2 $tpp->set( use_ixhash => 1 );
-
-This option keeps the order for every elements appeared in XML.
-L<Tie::IxHash> module is required.
-This makes parsing performance slow. (100% slower than default)
-
-=head2 $tpp->get( "option_name" );
-
-This method returns a current option value for "option_name".
+See also L</text_node_key> option.
 
 =head1 METHODS
 
-=head2 $tree = $tpp->parse( $source );
+=head2 new
+
+This constructor method returns a new XML::TreePP object with C<%options>.
+
+    $tpp = XML::TreePP->new( %options );
+
+=head2 set
+
+This method sets a option value for C<option_name>.
+If C<$option_value> is not defined, its option is deleted.
+
+    $tpp->set( option_name => $option_value );
+
+See OPTIONS section below for details.
+
+=head2 get
+
+This method returns a current option value for C<option_name>.
+
+    $tpp->get( 'option_name' );
+
+=head2 parse
 
 This method reads XML source and returns a hash tree converted.
 The first argument is a scalar or a reference to a scalar.
 
-=head2 $tree = $tpp->parsefile( $file );
+        $tree = $tpp->parse( $source );
+
+=head2 parsefile
 
 This method reads a XML file and returns a hash tree converted.
 The first argument is a filename.
 
-=head2 $tree = $tpp->parsehttp( $method, $url, $body, $head );
+    $tree = $tpp->parsefile( $file );
+
+=head2 parsehttp
 
 This method receives a XML file from a remote server via HTTP and
 returns a hash tree converted.
-$method is a method of HTTP connection: GET/POST/PUT/DELETE
-$url is an URI of a XML file.
-$body is a request body when you use POST method.
-$head is a request headers as a hash ref.
+
+    $tree = $tpp->parsehttp( $method, $url, $body, $head );
+
+C<$method> is a method of HTTP connection: GET/POST/PUT/DELETE
+C<$url> is an URI of a XML file.
+C<$body> is a request body when you use POST method.
+C<$head> is a request headers as a hash ref.
 L<LWP::UserAgent> module or L<HTTP::Lite> module is required to fetch a file.
 
-=head2 $source = $tpp->write( $tree, $encode );
+    ( $tree, $xml, $code ) = $tpp->parsehttp( $method, $url, $body, $head );
+
+In array context, This method returns also raw XML source received
+and HTTP response's status code.
+
+=head2 write
 
 This method parses a hash tree and returns a XML source generated.
-$tree is a referecen to a hash tree.
 
-=head2 $tpp->writefile( $file, $tree, $encode );
+    $source = $tpp->write( $tree, $encode );
+
+C<$tree> is a reference to a hash tree.
+
+=head2 writefile
 
 This method parses a hash tree and writes a XML source into a file.
-$file is a filename to create.
-$tree is a referecen to a hash tree.
+
+    $tpp->writefile( $file, $tree, $encode );
+
+C<$file> is a filename to create.
+C<$tree> is a reference to a hash tree.
+
+=head1 OPTIONS FOR PARSING XML
+
+This module accepts option parameters following:
+
+=head2 force_array
+
+This option allows you to specify a list of element names which
+should always be forced into an array representation.
+
+    $tpp->set( force_array => [ 'rdf:li', 'item', '-xmlns' ] );
+
+The default value is null, it means that context of the elements
+will determine to make array or to keep it scalar or hash.
+Note that the special wildcard name C<'*'> means all elements.
+
+=head2 force_hash
+
+This option allows you to specify a list of element names which
+should always be forced into an hash representation.
+
+    $tpp->set( force_hash => [ 'item', 'image' ] );
+
+The default value is null, it means that context of the elements
+will determine to make hash or to keep it scalar as a text node.
+See also L</text_node_key> option below.
+Note that the special wildcard name C<'*'> means all elements.
+
+=head2 cdata_scalar_ref
+
+This option allows you to convert a cdata section into a reference
+for scalar on parsing XML source.
+
+    $tpp->set( cdata_scalar_ref => 1 );
+
+The default value is false, it means that each cdata section is converted into a scalar.
+
+=head2 user_agent
+
+This option allows you to specify a HTTP_USER_AGENT string which
+is used by parsehttp() method.
+
+    $tpp->set( user_agent => 'Mozilla/4.0 (compatible; ...)' );
+
+The default string is C<'XML-TreePP/#.##'>, where C<'#.##'> is
+substituted with the version number of this library.
+
+=head2 http_lite
+
+This option forces pasrsehttp() method to use a L<HTTP::Lite> instance.
+
+    my $http = HTTP::Lite->new();
+    $tpp->set( http_lite => $http );
+
+=head2 lwp_useragent
+
+This option forces pasrsehttp() method to use a L<LWP::UserAgent> instance.
+
+    my $ua = LWP::UserAgent->new();
+    $ua->timeout( 60 );
+    $ua->env_proxy;
+    $tpp->set( lwp_useragent => $ua );
+
+You may use this with L<LWP::UserAgent::WithCache>.
+
+=head2 base_class
+
+This blesses class name for each element's hashref.
+Each class is named straight as a child class of it parent class.
+
+    $tpp->set( base_class => 'MyElement' );
+    my $xml  = '<root><parent><child key="val">text</child></parent></root>';
+    my $tree = $tpp->parse( $xml );
+    print ref $tree->{root}->{parent}->{child}, "\n";
+
+A hash for <child> element above is blessed to C<MyElement::root::parent::child>
+class. You may use this with L<Class::Accessor>.
+
+=head2 elem_class
+
+This blesses class name for each element's hashref.
+Each class is named horizontally under the direct child of C<MyElement>.
+
+    $tpp->set( base_class => 'MyElement' );
+    my $xml  = '<root><parent><child key="val">text</child></parent></root>';
+    my $tree = $tpp->parse( $xml );
+    print ref $tree->{root}->{parent}->{child}, "\n";
+
+A hash for <child> element above is blessed to C<MyElement::child> class.
+
+=head1 OPTIONS FOR WRITING XML
+
+=head2 first_out
+
+This option allows you to specify a list of element/attribute
+names which should always appears at first on output XML code.
+
+    $tpp->set( first_out => [ 'link', 'title', '-type' ] );
+
+The default value is null, it means alphabetical order is used.
+
+=head2 last_out
+
+This option allows you to specify a list of element/attribute
+names which should always appears at last on output XML code.
+
+    $tpp->set( last_out => [ 'items', 'item', 'entry' ] );
+
+=head2 indent
+
+This makes the output more human readable by indenting appropriately.
+
+    $tpp->set( indent => 2 );
+
+This doesn't strictly follow the XML Document Spec but does looks nice.
+
+=head2 xml_decl
+
+This module generates an XML declaration on writing an XML code per default.
+This option forces to change or leave it.
+
+    $tpp->set( xml_decl => '' );
+
+=head2 output_encoding
+
+This option allows you to specify a encoding of xml file generated
+by write/writefile methods.
+
+    $tpp->set( output_encoding => 'UTF-8' );
+
+On Perl 5.8.0 and later, you can select it from every
+encodings supported by Encode.pm. On Perl 5.6.x and before with
+Jcode.pm, you can use C<Shift_JIS>, C<EUC-JP>, C<ISO-2022-JP> and
+C<UTF-8>. The default value is C<UTF-8> which is recommended encoding.
+
+=head1 OPTIONS FOR BOTH
+
+=head2 utf8_flag
+
+This makes utf8 flag on for every element's value parsed
+and makes it on for an XML code generated as well.
+
+    $tpp->set( utf8_flag => 1 );
+
+Perl 5.8.1 or later is required to use this.
+
+=head2 attr_prefix
+
+This option allows you to specify a prefix character(s) which
+is inserted before each attribute names.
+
+    $tpp->set( attr_prefix => '@' );
+
+The default character is C<'-'>.
+Or set C<'@'> to access attribute values like E4X, ECMAScript for XML.
+Zero-length prefix C<''> is available as well, it means no prefix is added.
+
+=head2 text_node_key
+
+This option allows you to specify a hash key for text nodes.
+
+    $tpp->set( text_node_key => '#text' );
+
+The default key is C<#text>.
+
+=head2 ignore_error
+
+This module calls Carp::croak function on an error per default.
+This option makes all errors ignored and just return.
+
+    $tpp->set( ignore_error => 1 );
+
+=head2 use_ixhash
+
+This option keeps the order for each element appeared in XML.
+L<Tie::IxHash> module is required.
+
+    $tpp->set( use_ixhash => 1 );
+
+This makes parsing performance slow.
+(about 100% slower than default)
 
 =head1 AUTHOR
 
@@ -273,8 +395,8 @@ Yusuke Kawasaki, http://www.kawa.net/
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (c) 2006-2007 Yusuke Kawasaki. All rights reserved.
-This program is free software; you can redistribute it and/or 
+Copyright (c) 2006-2008 Yusuke Kawasaki. All rights reserved.
+This program is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
 
 =cut
@@ -285,7 +407,7 @@ use Carp;
 use Symbol;
 
 use vars qw( $VERSION );
-$VERSION = '0.21';
+$VERSION = '0.33';
 
 my $XML_ENCODING      = 'UTF-8';
 my $INTERNAL_ENCODING = 'UTF-8';
@@ -339,6 +461,9 @@ sub writefile {
     my $encode = shift;
     return $self->die( 'Invalid filename' ) unless defined $file;
     my $text = $self->write( $tree, $encode );
+    if ( $] >= 5.008001 && utf8::is_utf8( $text ) ) {
+        utf8::encode( $text );
+    }
     $self->write_raw_xml( $file, $text );
 }
 
@@ -364,20 +489,26 @@ sub write {
         $self->{__last_out} = { map { $keys->[$_] => $_ } 0 .. $#$keys };
     }
 
-    my $tnk = $self->{text_node_key};
-    local $self->{text_node_key} = $TEXT_NODE_KEY;
-    $self->{text_node_key} = $tnk if defined  $tnk;
+    my $tnk = $self->{text_node_key} if exists $self->{text_node_key};
+    $tnk = $TEXT_NODE_KEY unless defined $tnk;
+    local $self->{text_node_key} = $tnk;
 
-    my $apre = $self->{attr_prefix};
-    $apre = $ATTR_PREFIX unless defined  $apre;
+    my $apre = $self->{attr_prefix} if exists $self->{attr_prefix};
+    $apre = $ATTR_PREFIX unless defined $apre;
     local $self->{__attr_prefix_len} = length($apre);
     local $self->{__attr_prefix_rex} = defined $apre ? qr/^\Q$apre\E/s : undef;
 
+    local $self->{__indent};
+    if ( exists $self->{indent} && $self->{indent} ) {
+        $self->{__indent} = ' ' x $self->{indent};
+    }
+
     my $text = $self->hash_to_xml( undef, $tree );
     if ( $from && $to ) {
         my $stat = $self->encode_from_to( \$text, $from, $to );
         return $self->die( "Unsupported encoding: $to" ) unless $stat;
     }
+
     return $text if ( $decl eq '' );
     join( "\n", $decl, $text );
 }
@@ -484,9 +615,10 @@ sub parsehttp_lwp {
     }
     $req->content($body) if defined $body;
     my $res = $ua->request($req);
-    return unless $res->is_success();
+    my $code = $res->code();
     my $text = $res->content();
-    $self->parse( \$text );
+    my $tree = $self->parse( \$text ) if $res->is_success();
+    wantarray ? ( $tree, $text, $code ) : $tree;
 }
 
 sub parsehttp_lite {
@@ -510,9 +642,10 @@ sub parsehttp_lite {
         $http->add_req_header( 'User-Agent', $self->{__user_agent} );
     }
     $http->{content} = $body if defined $body;
-    $http->request($url) or return;
+    my $code = $http->request($url) or return;
     my $text = $http->body();
-    $self->parse( \$text );
+    my $tree = $self->parse( \$text );
+    wantarray ? ( $tree, $text, $code ) : $tree;
 }
 
 sub parsefile {
@@ -525,41 +658,58 @@ sub parsefile {
 
 sub parse {
     my $self = shift;
-    my $textref = ref $_[0] ? $_[0] : \$_[0];
-    return $self->die( 'Invalid XML source' ) if ( ref($textref) ne 'SCALAR' );
-    return $self->die( 'Null XML source' ) unless defined $$textref;
+    my $text = ref $_[0] ? ${$_[0]} : $_[0];
+    return $self->die( 'Null XML source' ) unless defined $text;
 
-    my $to = $self->{internal_encoding} || $INTERNAL_ENCODING;
-    if ($to) {
-        my $from = &xml_decl_encoding($textref);
-        if ($from) {
-            my $stat = $self->encode_from_to( $textref, $from, $to );
-            return $self->die( "Unsupported encoding: $from" ) unless $stat;
-        }
+    my $from = &xml_decl_encoding(\$text) || $XML_ENCODING;
+    my $to   = $self->{internal_encoding} || $INTERNAL_ENCODING;
+    if ( $from && $to ) {
+        my $stat = $self->encode_from_to( \$text, $from, $to );
+        return $self->die( "Unsupported encoding: $from" ) unless $stat;
     }
 
     local $self->{__force_array};
+    local $self->{__force_array_all};
     if ( exists $self->{force_array} ) {
         my $force = $self->{force_array};
         $force = [$force] unless ref $force;
         $self->{__force_array} = { map { $_ => 1 } @$force };
+        $self->{__force_array_all} = $self->{__force_array}->{'*'};
     }
 
-    my $tnk = $self->{text_node_key};
-    local $self->{text_node_key} = $TEXT_NODE_KEY;
-    $self->{text_node_key} = $tnk if defined  $tnk;
+    local $self->{__force_hash};
+    local $self->{__force_hash_all};
+    if ( exists $self->{force_hash} ) {
+        my $force = $self->{force_hash};
+        $force = [$force] unless ref $force;
+        $self->{__force_hash} = { map { $_ => 1 } @$force };
+        $self->{__force_hash_all} = $self->{__force_hash}->{'*'};
+    }
 
-    my $apre = $self->{attr_prefix};
-    local $self->{attr_prefix} = $ATTR_PREFIX;
-    $self->{attr_prefix} = $apre if defined  $apre;
+    my $tnk = $self->{text_node_key} if exists $self->{text_node_key};
+    $tnk = $TEXT_NODE_KEY unless defined $tnk;
+    local $self->{text_node_key} = $tnk;
+
+    my $apre = $self->{attr_prefix} if exists $self->{attr_prefix};
+    $apre = $ATTR_PREFIX unless defined $apre;
+    local $self->{attr_prefix} = $apre;
 
     if ( exists $self->{use_ixhash} && $self->{use_ixhash} ) {
         return $self->die( "Tie::IxHash is required." ) unless &load_tie_ixhash();
     }
 
-    my $flat = $self->xml_to_flat($textref);
-    my $tree = $self->flat_to_tree( $flat, '' );
-    wantarray ? ( $tree, $$textref ) : $tree;
+    my $flat  = $self->xml_to_flat(\$text);
+    my $class = $self->{base_class} if exists $self->{base_class};
+    my $tree  = $self->flat_to_tree( $flat, '', $class );
+    if ( ref $tree ) {
+        if ( defined $class ) {
+            bless( $tree, $class );
+        }
+        elsif ( exists $self->{elem_class} && $self->{elem_class} ) {
+            bless( $tree, $self->{elem_class} );
+        }
+    }
+    wantarray ? ( $tree, $text ) : $tree;
 }
 
 sub xml_to_flat {
@@ -591,6 +741,7 @@ sub xml_to_flat {
           )
           = ( $1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13 );
         if ( defined $ahead && $ahead =~ /\S/ ) {
+            $ahead =~ s/([^\040-\076])/sprintf("\\x%02X",ord($1))/eg;
             $self->warn( "Invalid string: [$ahead] before <$match>" );
         }
 
@@ -600,7 +751,7 @@ sub xml_to_flat {
                 $node->{endTag}++;
             }
             elsif ( $contElem =~ s#/$## ) {
-                # one line
+                $node->{emptyTag}++;
             }
             else {
                 $node->{startTag}++;
@@ -652,6 +803,7 @@ sub flat_to_tree {
     my $self   = shift;
     my $source = shift;
     my $parent = shift;
+    my $class  = shift;
     my $tree   = {};
     my $text   = [];
 
@@ -671,46 +823,83 @@ sub flat_to_tree {
             return $self->die( "Invalid tag sequence: <$parent></$name>" );
         }
         my $elem = $node->{attributes};
+        my $forcehash = $self->{__force_hash_all} || $self->{__force_hash}->{$name};
+        my $subclass;
+        if ( defined $class ) {
+            my $escname = $name;
+            $escname =~ s/\W/_/sg;
+            $subclass = $class.'::'.$escname;
+        }
         if ( $node->{startTag} ) {              # recursive call
-            my $child = $self->flat_to_tree( $source, $name );
-            if ( ref $elem && scalar keys %$elem ) {
-                if ( UNIVERSAL::isa( $child, "HASH" ) ) {
+            my $child = $self->flat_to_tree( $source, $name, $subclass );
+            next unless defined $child;
+            my $hasattr = scalar keys %$elem if ref $elem;
+            if ( UNIVERSAL::isa( $child, "HASH" ) ) {
+                if ( $hasattr ) {
                     # some attributes and some child nodes
-                    foreach my $key ( keys %$child ) {
-                        $elem->{$key} = $child->{$key};
-                    }
+                    %$elem = ( %$elem, %$child );
                 }
-                elsif ( defined $child ) {
-                    # some attributes and text node
-                    $elem->{$self->{text_node_key}} = $child;
+                else {
+                    # some child nodes without attributes
+                    $elem = $child;
                 }
             }
             else {
-                # no attributes and text node or nothing
-                $elem = $child;
+                if ( $hasattr ) {
+                    # some attributes and text node
+                    $elem->{$self->{text_node_key}} = $child;
+                }
+                elsif ( $forcehash ) {
+                    # only text node without attributes
+                    $elem = { $self->{text_node_key} => $child };
+                }
+                else {
+                    # text node without attributes
+                    $elem = $child;
+                }
+            }
+        }
+        elsif ( $forcehash && ! ref $elem ) {
+            $elem = {};
+        }
+        # bless to a class by base_class or elem_class
+        if ( ref $elem && UNIVERSAL::isa( $elem, "HASH" ) ) {
+            if ( defined $subclass ) {
+                bless( $elem, $subclass );
+            } elsif ( exists $self->{elem_class} && $self->{elem_class} ) {
+                my $escname = $name;
+                $escname =~ s/\W/_/sg;
+                my $elmclass = $self->{elem_class}.'::'.$escname;
+                bless( $elem, $elmclass );
             }
         }
         # next unless defined $elem;
         $tree->{$name} ||= [];
         push( @{ $tree->{$name} }, $elem );
     }
-    foreach my $key ( keys %$tree ) {
-        next if $self->{__force_array}->{$key};
-        next if ( 1 < scalar @{ $tree->{$key} } );
-        $tree->{$key} = shift @{ $tree->{$key} };
+    if ( ! $self->{__force_array_all} ) {
+        foreach my $key ( keys %$tree ) {
+            next if $self->{__force_array}->{$key};
+            next if ( 1 < scalar @{ $tree->{$key} } );
+            $tree->{$key} = shift @{ $tree->{$key} };
+        }
     }
+    my $haschild = scalar keys %$tree;
     if ( scalar @$text ) {
         if ( scalar @$text == 1 ) {
+            # one text node (normal)
             $text = shift @$text;
         }
         elsif ( ! scalar grep {ref $_} @$text ) {
+            # some text node splitted
             $text = join( '', @$text );
         }
         else {
+            # some cdata node
             my $join = join( '', map {ref $_ ? $$_ : $_} @$text );
             $text = \$join;
         }
-        if ( scalar keys %$tree ) {
+        if ( $haschild ) {
             # some child nodes and also text node
             $tree->{$self->{text_node_key}} = $text;
         }
@@ -719,6 +908,10 @@ sub flat_to_tree {
             $tree = $text;
         }
     }
+    elsif ( ! $haschild ) {
+        # no child and no text
+        $tree = "";
+    }
     $tree;
 }
 
@@ -741,9 +934,12 @@ sub hash_to_xml {
     my $prelen = $self->{__attr_prefix_len};
     my $pregex = $self->{__attr_prefix_rex};
 
-    foreach my $loopkey ( $firstkeys, $allkeys, $lastkeys ) {
-        next unless ref $loopkey;
-        foreach my $key ( grep { ! $prelen || $_ !~ $pregex } @$loopkey ) {
+    foreach my $keys ( $firstkeys, $allkeys, $lastkeys ) {
+        next unless ref $keys;
+        my $elemkey = $prelen ? [ grep { $_ !~ $pregex } @$keys ] : $keys;
+        my $attrkey = $prelen ? [ grep { $_ =~ $pregex } @$keys ] : [];
+
+        foreach my $key ( @$elemkey ) {
             my $val = $hash->{$key};
             if ( !defined $val ) {
                 push( @$out, "<$key />" );
@@ -765,8 +961,8 @@ sub hash_to_xml {
                 push( @$out, $child );
             }
         }
-    
-        foreach my $key ( grep { $prelen && $_ =~ $pregex } @$loopkey ) {
+
+        foreach my $key ( @$attrkey ) {
             my $name = substr( $key, $prelen );
             my $val = &xml_escape( $hash->{$key} );
             push( @$attr, ' ' . $name . '="' . $val . '"' );
@@ -774,7 +970,14 @@ sub hash_to_xml {
     }
     my $jattr = join( '', @$attr );
 
-    # s/^(\s*<)/  $1/mg foreach @$out;              # indent
+    if ( defined $name && scalar @$out && ! grep { ! /^</s } @$out ) {
+        # Use human-friendly white spacing
+        if ( defined $self->{__indent} ) {
+            s/^(\s*<)/$self->{__indent}$1/mg foreach @$out;
+        }
+        unshift( @$out, "\n" );
+    }
+
     my $text = join( '', @$out );
     if ( defined $name ) {
         if ( scalar @$out ) {
@@ -814,7 +1017,6 @@ sub array_to_xml {
         }
     }
 
-    # s/^(\s*<)/  $1/mg foreach @$out;              # indent
     my $text = join( '', @$out );
     $text;
 }
@@ -823,7 +1025,9 @@ sub scalaref_to_cdata {
     my $self = shift;
     my $name = shift;
     my $ref  = shift;
-    my $text = '<![CDATA[' . $$ref . ']]>';
+    my $data = defined $$ref ? $$ref : '';
+    $data =~ s#(]])(>)#$1]]><![CDATA[$2#g;
+    my $text = '<![CDATA[' . $data . ']]>';
     $text = "<$name>$text</$name>\n" if ( $name ne $self->{text_node_key} );
     $text;
 }
@@ -861,8 +1065,10 @@ sub read_raw_xml {
 sub xml_decl_encoding {
     my $textref = shift;
     return unless defined $$textref;
-    my $args    = ( $$textref =~ /^\s*<\?xml(\s+\S.*)\?>/s )[0] or return;
-    my $getcode = ( $args =~ /\s+encoding="(.*?)"/ )[0] or return;
+    my $args    = ( $$textref =~ /^(?:\s*\xEF\xBB\xBF)?\s*<\?xml(\s+\S.*)\?>/s )[0] or return;
+    my $getcode = ( $args =~ /\s+encoding=(".*?"|'.*?')/ )[0] or return;
+    $getcode =~ s/^['"]//;
+    $getcode =~ s/['"]$//;
     $getcode;
 }
 
@@ -871,17 +1077,38 @@ sub encode_from_to {
     my $txtref = shift or return;
     my $from   = shift or return;
     my $to     = shift or return;
-    return $to if ( uc($from) eq uc($to) );
-    &load_encode() if ( $] > 5.008 );
 
     unless ( defined $Encode::EUCJPMS::VERSION ) {
         $from = 'EUC-JP' if ( $from =~ /\beuc-?jp-?(win|ms)$/i );
         $to   = 'EUC-JP' if ( $to   =~ /\beuc-?jp-?(win|ms)$/i );
     }
 
-    if ( defined $Encode::VERSION ) {
+    if ( $from =~ /^utf-?8$/i ) {
+        $$txtref =~ s/^\xEF\xBB\xBF//s;         # UTF-8 BOM (Byte Order Mark)
+    }
+
+    my $setflag = $self->{utf8_flag} if exists $self->{utf8_flag};
+    if ( $] < 5.008001 && $setflag ) {
+        return $self->die( "Perl 5.8.1 is required for utf8_flag: $]" );
+    }
+
+    if ( $] >= 5.008 ) {
+        &load_encode();
         my $check = ( $Encode::VERSION < 2.13 ) ? 0x400 : Encode::FB_XMLCREF();
-        Encode::from_to( $$txtref, $from, $to, $check );
+        if ( $] >= 5.008001 && utf8::is_utf8( $$txtref ) ) {
+            if ( $to =~ /^utf-?8$/i ) {
+                # skip
+            } else {
+                $$txtref = Encode::encode( $to, $$txtref, $check );
+            }
+        } else {
+            $$txtref = Encode::decode( $from, $$txtref );
+            if ( $to =~ /^utf-?8$/i && $setflag ) {
+                # skip
+            } else {
+                $$txtref = Encode::encode( $to, $$txtref, $check );
+            }
+        }
     }
     elsif ( (  uc($from) eq 'ISO-8859-1'
             || uc($from) eq 'US-ASCII'
@@ -949,6 +1176,7 @@ sub get_jcode_name {
 
 sub xml_escape {
     my $str = shift;
+    return '' unless defined $str;
     # except for TAB(\x09),CR(\x0D),LF(\x0A)
     $str =~ s{
         ([\x00-\x08\x0B\x0C\x0E-\x1F\x7F])
